#!/usr/bin/env python
# -*- coding: latin-1 -*-
r"""
__   __         _    _ _
\ \ / /__ _ __ | | _(_) |_
 \ V / _ \ '_ \| |/ / | __|
  | |  __/ |_) |   <| | |_
  |_|\___| .__/|_|\_\_|\__|
         |_|

Yepkit firmware update tool v0.2.0
http://yepkit.com/

Includes PyUSB
http://walac.github.io/pyusb/

Copyright 2015, 2013 Yepkit Lda and other contributors
Released under the MIT license, please read the file LICENSE.txt
http://yepkit.com/mit_license

Date: 2015-10-17

Usage:
  ykfirmware (-v | --vendor [VENDOR ID]) (-p | --product [PRODUCT ID]) (-s | --serial [SERIAL]) [HEX FILENAME]
  joe (-h | --help)
  joe --version
Options:
  -h --help     Show this screen.
  --version     Show version.

Notes:
  * None

"""

import sys
import argparse
import time
import struct
import ctypes
import array
import binascii

__version__ = '0.2.0'

# Default (ok, our first...) public updatable product
PYK24_USB_VENDOR_ID = 0x04d8
PYK24_USB_PRODUCT_ID = 0xf2f7
### BOOTLOADER: 0xf2f7
### DEMO APPS: 0xf25c

# Bootloader specific consts
YKFW_USB_TIMEOUT = 333  # timeout in ms
YKFW_USB_INTERFACE = 0
YKFW_USB_EP_IN_ADDR = 0x01
YKFW_USB_EP_OUT_ADDR = 0x81
YKFW_USB_PACKET_SIZE = 64

# Bootloader packet commands
# We invite you to take a look at the bootloader source code itself if you are interested in additional functions
YKFW_CMD_QUERY_DEVICE = 0x02
YKFW_CMD_UNLOCK_CONFIG = 0x03
YKFW_CMD_ERASE_DEVICE = 0x04
YKFW_CMD_PROGRAM_DEVICE = 0x05
YKFW_CMD_PROGRAM_COMPLETE = 0x06
YKFW_CMD_GET_DATA = 0x07
YKFW_CMD_RESET_DEVICE = 0x08
YKFW_CMD_SIGN_FLASH = 0x09
YKFW_CMD_QUERY_EXTENDED_INFO = 0x0C


# Not the most pythonic way to print to the terminal but we do prefer it to avoid weird future imports, flush or
# compatibility issues
def printout(str='', end='\n', flush=True):
    sys.stdout.write(str + end)
    if flush:
        sys.stdout.flush()


# Same as the previous but pointing to stderr
def printerr(str='', end='\n', flush=True):
    sys.stderr.write(str + end)
    if flush:  # a paranoid touch as stderr is supposed to be non buffered by default
        sys.stderr.flush()


# The cool stuff starts here
def main():

    # Attempt to import pyusb
    try:
        import usb.core
        import usb.util
    except ImportError:
        printerr('> Error attempting to import pyusb, please validate pyusb is installed.')
        if e.message is not None and e.message != '':
            printerr('> Error message: %s' % e.message)
        return 1

    # Parser definition
    parser = argparse.ArgumentParser(description='Yepkit firmware update tool.')
    parser.add_argument('infile', type=argparse.FileType('r'), help='the input Intel HEX filename')
    parser.add_argument('-v', '--vendor', type=int, default=PYK24_USB_VENDOR_ID, help='the USB vendor ID')
    parser.add_argument('-p', '--product', type=int, default=PYK24_USB_PRODUCT_ID, help='the USB product ID')
    parser.add_argument('-s', '--serial', default='', help='the USB device serial number string descriptor')
    args = parser.parse_args()
    dev = None
    progressref = 0
    anotherserial = ''
    printout('Welcome to %s\n%s' % (parser.description, 'Progress status (8 steps):'))
    try:
        while True:
            for dev in usb.core.find(find_all=True, idVendor=args.vendor, idProduct=args.product):
                if isinstance(dev, usb.core.Device):
                    if args.serial == '' or args.serial == dev.serial_number:
                        if progressref > 3:
                            printout()
                        break
                    anotherserial = dev.serial_number
                dev = None
            if dev is not None:
                break
            else:
                if progressref == 3:
                    printout('> Specified device not found in bootloader mode, please reset it now or\n'
                             '> disconnect and reconnect it promptly.')
                    printout('> Note: the tool needs to access a bootloader function that is only\n'
                             '>       available in the early device startup.')
                    printout('> Waiting for device', end='')
                elif progressref > 3 and progressref % 3 == 0:
                    printout('.', end='')
                if progressref > (15/.2):  # wait ~15s before give up
                    printout('\n> Tool timeout, please check the arguments and the specific IDs of your device.')
                    printout('> The tool is currently looking for:\n'
                             '>   USB (vendor:product) ID: %.4x:%.4x' % (args.vendor, args.product))
                    if args.serial != '':
                        printout('>   USB device serial: %s' % args.serial)
                    if anotherserial != '':
                        printout('>   WARNING, located the same device with another serial: %s' % anotherserial)
                    return 1
                time.sleep(.2)  # The bootloader timeup is about 2s, the lookup cycle must be fast
                progressref += 1
        printout('1. Device located.')
    except usb.core.USBError as e:
        printerr('> Error attempting to use pyusb, please validate libusb is correctly installed on your computer.')
        if e.message is not None and e.message != '':
            printerr('> Error message: %s' % e.message)
        return 1

    # Detach kernel driver and set device configuration
    try:
        if dev.is_kernel_driver_active(YKFW_USB_INTERFACE):
            try:
                dev.detach_kernel_driver(YKFW_USB_INTERFACE)
            except usb.core.USBError as e:
                if e.message is not None and e.message != '':
                    printerr('> Got an error attempting to detach the kernel driver: %s' % e.message)
        dev.set_configuration()
        usb.util.claim_interface(dev, YKFW_USB_INTERFACE)
        time.sleep(.3)  # Settle down a bit
        printout('2. Device successfully connected.')
    except usb.core.USBError as e:
        printerr('> Could not connect to the device, please check your user permissions.')
        if e.message is not None and e.message != '':
            printerr('> Error message: %s' % e.message)
        return 1

    # Query the device bootloader settings
    devbytesperaddress, devmemaddress, devmemlength = 0, 0, 0
    attempt = 0
    sendpacket = struct.pack('<B', YKFW_CMD_QUERY_DEVICE)
    sendpacket += ctypes.create_string_buffer(YKFW_USB_PACKET_SIZE - len(sendpacket)).raw
    while True:
        try:
            attempt += 1
            sz = dev.write(YKFW_USB_EP_IN_ADDR, sendpacket, YKFW_USB_TIMEOUT)
            recvpacket = dev.read(YKFW_USB_EP_OUT_ADDR, YKFW_USB_PACKET_SIZE, YKFW_USB_TIMEOUT)
            (bogus, devpacketdatafieldsize, devbytesperaddress, devmemtype, devmemaddress, devmemlength) = \
                struct.unpack_from('<4B2L', recvpacket)
            printout('3. Device programmable region: 0x%x to 0x%x, a total of %iKiB.' % (devmemaddress, devmemaddress +
                                                                                         devmemlength, devmemlength *
                                                                                         devbytesperaddress / 1024))
            break
        except usb.core.USBError as e:
            if attempt < 3:
                time.sleep(.3)
            else:
                printerr('> Got an error trying to query the device.')
                if e.message is not None and e.message != '':
                    printerr('> Error message: %s' % e.message)
                return 1

    # Parse the Intel HEX file
    printout('4. Importing .hex file...', end='')
    hexrecinfo = struct.Struct('>BHB')  # Big-endian modifier and the formatting structs
    hexrecsegm = struct.Struct('>H')
    flashbuffer = [0xffff for i in range(devmemlength)]  ## // devbytesperaddress)]  # Populate the flash buffer
    ln = 0  # Line number reference
    segmaddr = 0  # Segment address reference
    for rec in args.infile.readlines():  # For each .hex file line
        ln += 1  # Keep the line number info in case we need to report an error
        if rec[0] != ':':  # Minimalistic consistency checks
            printout()
            printerr('> Unrecognized Intel HEX format, affected line: %i' % ln)
            return 1
        (rsize, raddr, rtype) = hexrecinfo.unpack(binascii.unhexlify(rec[1:9]))  # Get the record size, address and type
        raddr += segmaddr
        rsum = 0  # Sum the record bytes according to the proposed format
        for i in range(1, 9 + rsize * 2, 2):
            rsum += int(rec[i:(i + 2)], 16)
        csum = (~rsum + 1) & 0xff  # Compute the checksum based on the two's complement and ensure the result is a byte
        if int(rec[(9 + rsize * 2):(11 + rsize * 2)], 16) != csum:  # A few more validations before proceed
            # Oops, checksum mismatch
            printout()
            printerr('> Checksum mismatch, affected line: %i' % ln)
            return 1
        elif rsize % devbytesperaddress:
            # Unaligned records
            printout()
            printerr('> Expecting a multiple of %i byte(s) record, affected line: %i' % ln)
            return 1
        if rtype in (0x02, 0x04):
            # Segment or linear addresses
            segmaddr = hexrecsegm.unpack(binascii.unhexlify(rec[9:13]))[0]
            if rtype == 0x02:
                segmaddr <<= 4
            else:
                segmaddr <<= 16
        elif rtype == 0:
            # Data type record
            devaddr = raddr // devbytesperaddress
            rpayload = array.array('H', binascii.unhexlify(rec[9:(9 + rsize * 2)]))  # Get the payload
            mstart, mend = int(devaddr - devmemaddress), int(devaddr + len(rpayload) - devmemaddress)
            # Fill the buffer if between boundaries
            if 0 <= mstart < devmemaddress + devmemlength:
                if mend > devmemaddress + devmemlength:  # Hummm this does not smell good
                    mend = devmemaddress + devmemlength
                flashbuffer[mstart:mend] = rpayload[:mend - mstart]
    printout('done.')

    # Erase device
    printout('5. Erasing device before programming...', end='')
    try:
        sendpacket = struct.pack('<B', YKFW_CMD_ERASE_DEVICE)
        sendpacket += ctypes.create_string_buffer(YKFW_USB_PACKET_SIZE - len(sendpacket)).raw
        sz = dev.write(YKFW_USB_EP_IN_ADDR, sendpacket, YKFW_USB_TIMEOUT)
    except usb.core.USBError as e:
        printout()
        printerr('> Got an error trying to erase the device.')
        if e.message is not None and e.message != '':
            printerr('> Error message: %s', e.message)
        return 1
    printout('done.')

    # Program device
    printout('6. Programming device...', end='')
    offset = devmemaddress
    for chunk in (flashbuffer[pos:pos + 56 // devbytesperaddress] for pos in
                  range(0, len(flashbuffer), 56 // devbytesperaddress)):
        if min(chunk) < 0xffff:  # It's safe to skip 0xffff filled chunks
            sendpacket = struct.pack('<BLBBB%iH' % len(chunk), YKFW_CMD_PROGRAM_DEVICE, offset,
                                     len(chunk) * devbytesperaddress, 0, 0, *chunk)
            if len(sendpacket) < YKFW_USB_PACKET_SIZE:
                sendpacket += ctypes.create_string_buffer(YKFW_USB_PACKET_SIZE - len(sendpacket)).raw
            attempt = 0
            while True:
                attempt += 1
                try:
                    sz = dev.write(YKFW_USB_EP_IN_ADDR, sendpacket, YKFW_USB_TIMEOUT)
                    time.sleep(.01)
                    break
                except usb.core.USBError as e:
                    time.sleep(.17)
                    if attempt == 17:
                        printout()
                        printerr('> Got an error trying to program/flash the device at address: 0x%x' % offset)
                        if e.message is not None and e.message != '':
                            printerr('> Error message: %s' % e.message)
                        return 1
        offset += 56 // devbytesperaddress

    # Finish signalling program complete
    try:
        sendpacket = struct.pack('<B', YKFW_CMD_PROGRAM_COMPLETE)
        sendpacket += ctypes.create_string_buffer(YKFW_USB_PACKET_SIZE - len(sendpacket)).raw
        sz = dev.write(YKFW_USB_EP_IN_ADDR, sendpacket, YKFW_USB_TIMEOUT)
    except usb.core.USBError as e:
        printout()
        printerr('> Got an error trying to complete the device programming.')
        if e.message is not None and e.message != '':
            printerr('> Error message: %s' % e.message)
        return 1
    printout('done.')

    # Verify the written flash
    printout('7. Verifying the written data...', end='')
    offset = devmemaddress
    for chunk in (flashbuffer[pos:pos + 56 // devbytesperaddress] for pos in
                  range(0, len(flashbuffer), 56 // devbytesperaddress)):
        sendpacket = struct.pack('<BLB', YKFW_CMD_GET_DATA, offset, len(chunk) * devbytesperaddress)
        if len(sendpacket) < YKFW_USB_PACKET_SIZE:
            sendpacket += ctypes.create_string_buffer(YKFW_USB_PACKET_SIZE - len(sendpacket)).raw
        attempt = 0
        while True:
            attempt += 1
            try:
                sz = dev.write(YKFW_USB_EP_IN_ADDR, sendpacket, YKFW_USB_TIMEOUT)
                recvpacket = dev.read(YKFW_USB_EP_OUT_ADDR, YKFW_USB_PACKET_SIZE, YKFW_USB_TIMEOUT)
                recvunpacked = struct.unpack_from('<BHBBBBB%iH' % ((len(recvpacket) - 8) // devbytesperaddress), recvpacket)
                for i, j in zip(chunk, recvunpacked[7:]):
                    if i != j and i != 0xffff and j != 0xff:  # Obvious comparison plus a lazy ghost byte discard
                        printout()
                        printerr('> Data inconsistency detected at the offset: 0x%x.' % offset)
                        return 1
                break  # Comparison complete, leave the loop
            except usb.core.USBError as e:
                time.sleep(.1)
                if attempt == 3:
                    printout()
                    printerr('> Got an error trying to retrieve the device data at address: 0x%x' % offset)
                    if e.message is not None and e.message != '':
                        printerr('> Error message: %s' % e.message)
                    return 1
        offset += 56 // devbytesperaddress
    printout('done.')

    # Finally reset the device to start the user application, good luck with the tests! :)
    printout('8. Resetting device...', end='')
    try:
        sendpacket = struct.pack('<B', YKFW_CMD_RESET_DEVICE)
        sendpacket += ctypes.create_string_buffer(YKFW_USB_PACKET_SIZE - len(sendpacket)).raw
        sz = dev.write(YKFW_USB_EP_IN_ADDR, sendpacket, YKFW_USB_TIMEOUT)
        usb.util.dispose_resources(dev)
        dev = None
    except usb.core.USBError as e:
        printout()
        printerr('> Got an error trying to reset the device.')
        if e.message is not None and e.message != '':
            printerr('> Error message: %s' % e.message)
        return 1
    printout('done.')


if __name__ == '__main__':
    main()
